using System;
using System.Collections.Generic;
using M17N;
using M17N.Core;

public class Test
{
  const int LENGTH = 9;
  static MText mt = new MText ("012345678");
  static MSymbol key = MSymbol.PropertyKey ("fse",
					    MProperty.Flags.FrontSensitive);
  static MSymbol val0 = MSymbol.Of ("0");
  static MSymbol val1 = MSymbol.Of ("1");
  static MSymbol val2 = MSymbol.Of ("2");
  static MProperty prop0 = new MProperty (key, val0);
  static MProperty prop1 = new MProperty (key, val1);
  static MProperty prop2 = new MProperty (key, val2);

  // Each property is identified by its beginning point.
  // The character at position i has the property that begins at prop[i].
  // If the character has no property, prop[i] == -1.
  static int[] prop = new int[LENGTH];

  // The property beginning at position i ends at position end[i].
  // If no property begins at position i, end[i] is undefined.
  static int[] end = new int[LENGTH];

  // The property beginning at position i has value value[i].
  // If no property begings at positoin i, value[i] is undefined.
  static MSymbol[] value = new MSymbol[LENGTH];

  // Cut buffers
  static int[] subprop = new int[LENGTH], subend = new int[LENGTH];
  static MSymbol[] subvalue = new MSymbol[LENGTH];

  static void TestPushProp (int from, int to, MProperty newprop)
  {
    Console.WriteLine ("from {0}, to {1}, push {2}.\n", from, to, newprop.Val);

    int i, n;

    if (from > 0 && prop[from - 1] == prop[from] && prop[from] != -1)
      {
	n = end[prop[from]];
	for (i = prop[from]; i < n; i++)
	  prop[i] = -1;
      }
    if (to < LENGTH && prop[to - 1] == prop[to] && prop[to] != -1)
      {
	n = end[prop[to]];
	for (i = to; i < n; i++)
	  prop[i] = -1;
      }
    for (i = from; i < to; i++)
      prop[i] = from;
    end[from] = to;
    value[from] = (MSymbol) newprop.Val;

    mt.PushProp (from, to, newprop);
  }

  static void TestPopProp (int from, int to, MSymbol key)
  {
    Console.WriteLine ("from {0}, to {1}, pop.\n", from, to);

    int i, n;

    if (from > 0 && prop[from - 1] == prop[from] && prop[from] != -1)
      for (i = prop[from]; i < from; i++)
	  prop[i] = -1;

    if (to < LENGTH && prop[to - 1] == prop[to] && prop[to] != -1)
      {
	n = end[prop[to]];
	for (i = to; i < n; i++)
	  prop[i] = -1;
      }

    for (i = from; i < to; i++)
      prop[i] = -1;

    mt.PopProp (from, to, key);
  }
    
  static void TestDelIns (int from, int to, int from2)
  {
    Console.WriteLine ("from {0}, to {1}, moveto {2}.\n", from, to, from2);

    int i, n, l = to - from;

    // sensitivity for deletion
    if (from > 0 && prop[from] != -1)
      {
	n = end[prop[from]];
	for (i = prop[from - 1] == prop[from] ? prop[from] : from; i < n; i++)
	  prop[i] = -1;
      }

    if (to < LENGTH && prop[to] != -1)
      {
	n = end[prop[to]];
	for (i = prop[to - 1] == prop[to] ? prop[to] : to; i < n; i++)
	  prop[i] = -1;
      }

    // copy
    for (i = from; i < to; i++)
      {
	if (prop[i] != -1)
	  {
	    subprop[i - from] = prop[i] - from + from2;
	    subend[i - from] = end[i] - from + from2;
	    subvalue[i - from] = value[i];
	  }
	else
	  subprop[i - from] = -1;
      }

    // delete
    for (i = to; i < LENGTH; i++)
      {
	if (prop[i] != -1)
	  {
	    prop[i - l] = prop[i] - l;
	    end[i - l] = end[i] - l;
	    value[i - l] = value[i];
	  }
	else
	  prop[i - l] = -1;
      }
    prop[LENGTH - l] = -1;

    // sensitivity for insertion
    if (prop[from2] != -1)
      {
	n = end[prop[from2]];
	if (from2 > 0 && prop[from2 - 1] == prop[from2])
	  i = prop[from2];
	else
	  i = from2;
	for (; i < n; i++)
	  prop[i] = -1;
      }
    if (from2 > 0 && subprop[0] != -1)
      for (i = 0; i < subend[0] - from2; i++)
	subprop[i] = -1;

    // move
    for (i = LENGTH - 1; i >= from2 + l; i--)
      {
	if (prop[i - l] != -1)
	  {
	    prop[i] = prop[i - l] + l;
	    end[i] = end[i - l] + l;
	    value[i] = value[i - l];
	  }
	else
	  prop[i] = -1;
      }

    // insert
    for (i = from2; i < from2 + l; i++)
      {
	if (subprop[i - from2] != -1)
	  {
	    prop[i] = subprop[i - from2];
	    end[i] = subend[i - from2];
	    value[i] = subvalue[i - from2];
	  }
	else
	  prop[i] = -1;
      }

    MText mt2 = mt.Dup ();
    mt.Del (from, to);
    mt.Ins (from2, mt2, from, to);
  }

  static bool Compare ()
  {
    for (int i = 0; i < LENGTH; i++)
      {
	object val = mt.GetProp (i, key);

	if (prop[i] == -1 && (MSymbol) val != null)
	  {
	    Console.WriteLine ("value[{0}] is nil, GetProp returned {1}",
			       i,
			       val == null ? MSymbol.nil : val);
	    return false;
	  }

	if (prop[i] != -1 && (value[prop[i]] != (MSymbol) val))
	  {
	    Console.WriteLine ("value[{0}] is {1}, GetProp returned {2}",
			       i,
			       value[prop[i]] == null ? MSymbol.nil : value[prop[i]],
			       val == null ? MSymbol.nil : val);
	    return false;
	  }
      }
    return true;
  }

  static void Dump ()
  {
    for (int i = 0; i <= LENGTH; i++)
      Console.Write ("{0} ", i);
    Console.WriteLine ("\n-------------------");

    if (prop[0] == -1)
      Console.Write ("  ");
    else
      Console.Write ("{0} ", value[prop[0]]);

    for (int i = 1; i < LENGTH; i++)
      {
	if (prop[i] == -1)
	  Console.Write ("  ");
	else if (prop[i - 1] == prop[i])
	  Console.Write ("< ");
	else
	  Console.Write ("{0} ", value[prop[i]]);
      }
    Console.Write ("\n");
  }

  public static void Main (string[] args)
  {
    Random r = new Random (int.Parse (args[0]));
    int check = (args.Length > 1 ? int.Parse (args[1]) : 0xFFFFFFF);

    for (int i = 0; i < LENGTH; i++)
      prop[i] = -1;	

    for (int loop = 0; loop < 1000000; loop++)
      {
	int from, to;

	Console.WriteLine ("--- loop = {0} ---\n", loop);
	if (loop >= check)
	  {
	    mt.DumpPropNested ();
	    Dump ();
	    M17n.debug = true;
	  }

	do {
	  from = r.Next (LENGTH);
	  to = r.Next (LENGTH) + 1;
	} while (from == to);
	if (from > to)
	  {
	    int tmp = from;
	    from = to;
	    to = tmp;
	  }

	MProperty prop;
	switch (r.Next (3))
	  {
	  case 0:
	    prop = prop0;
	    break;
	  case 1:
	    prop = prop1;
	    break;
	  default:
	    prop = prop2;
	    break;
	  }

	switch (r.Next (3))
	  {
	  case 0:
	    TestPushProp (from, to, prop);
	    break;
	  case 1:
	    TestPopProp (from, to, key);
	    break;
	  case 2:
	    TestDelIns (from, to, r.Next (LENGTH - (to - from) + 1));
	    break;
	  }

	if (M17n.debug)
	  mt.DumpPropNested ();

	if (Compare () == false)
	  {
	    Console.WriteLine ("");
	    Dump ();
	    Console.WriteLine ("Failed.");
	    return;
	  }
      }
  }
}
